// use crate::get_client::RoClient;
// use color_eyre::eyre::Result;
// use crate::ls::coder_channel::DecodeChannel;
//
// pub async fn ls_tree(client: &RoClient, path: &str, level: u16, full: bool, coder: DecodeChannel) -> Result<()> {
//     todo!()
// }
