//! Definitions of associated methods for response objects.
mod downloadable;
mod feed;
mod pipeline;
mod plugin;
mod plugininstance;

pub use downloadable::*;
pub use feed::*;
pub use pipeline::*;
pub use plugin::*;
pub use plugininstance::*;
