//! Unicode symbols
//!
//! https://en.wikipedia.org/wiki/Geometric_Shapes_(Unicode_block)

pub const WHITE_CIRCLE: &str = "\u{25CB}";
pub const DOTTED_CIRCLE: &str = "\u{25CC}";
pub const BLACK_CIRCLE: &str = "\u{25CF}";

pub const BLACK_UP_POINTING_TRIANGLE: &str = "\u{25B2}";
pub const BLACK_DOWN_POINTING_TRIANGLE: &str = "\u{25BC}";

pub const HORIZONTAL_BAR: &str = "\u{2015}";

pub const VERTICAL_BAR: &str = "\u{007C}";

pub const CHECK_MARK: &str = "\u{2713}";
